/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#ifndef OAPR_POOL_HPP
#define OAPR_POOL_HPP
#include "oapr/oapr.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include <string>
#include "apr_pools.h"
#include "oapr/oapr_exception.hpp"
#include "boost/shared_ptr.hpp"
FILE_VERSION("$Revision: 1.6 $",oapr_pool_hpp);

namespace oapr
{
  class LIBOAPR_API pool
  {
  public:
    typedef boost::shared_ptr<pool> pool_ptr;
  public:
    pool()
      : m_stat(0)
	,m_id("")
	,m_dtor(true)
    {
      OAPR_CHECK(apr_pool_create(&m_pool,NULL));
    }
    /**
     * construct a pool in the category cat
     * @param id the id for this pool
     */
    pool(const std::string& id)
      : m_pool(0)
      , m_stat(0)
      , m_id(id)
      ,m_dtor(true){
      OAPR_CHECK(apr_pool_create(&m_pool,NULL));
    }
    /**
     * place a raw apr pool pointer into a pool.
     * This will allow you to assign a pool to an
     * oapr object and determine what will happen to
     * pool when the object dtors
     * @param p The raw apr pool
     * @param id The id to assign to the pool
     * @param dtor Whether to destroy the pool with this dtors
     */
    pool(apr_pool_t* p, const std::string &id,bool dtor)
      : m_pool(p)
	, m_stat(0)
	, m_id(id)
	, m_dtor(dtor)
    {
    }

    /**
     * create a subpool of p in category cat
     * @param p the pool to create a subpool from
     * @param id the id for this pool
     */
    pool(apr_pool_t* p, const std::string& id)
      : m_pool(0), m_stat(0),m_id(id),m_dtor(true){
      OAPR_CHECK(apr_pool_create(&m_pool,p));
    }

    /**
     * destroy the pool, unless set otherwise
     */
    ~pool(){if(m_dtor){apr_pool_destroy(m_pool);}}

    /**
     * clears the pool. This does *not* free the memory. However,
     * it will destroy the object(s) created from the pool
     */
    void clear(){ apr_pool_clear(m_pool); }

    /**
     * for various reasons, destory should not be automatically
     * called on dtor of this pool. Call this function to set that
     * flag
     */
    void undtor(){m_dtor=false;}
    void dtor(){m_dtor=true;}

    /**
     * forcibly destroy this pool. 
     * @warning only call this if you absolutely know what you are doing.
     */
    void destroy(){
      m_dtor=false;
      apr_pool_destroy(m_pool);
    }

    /**
     * get the status of the pool, as in the apr_status_t 
     * set on ctor
     */
    apr_status_t stat()const{return m_stat;}

    /**
     * get the category that this pool is assigned to.
     * @return the category
     */
    std::string id()const{return m_id;}

    /**
     * get the  pool
     * @return pointer to an apr_pool_t
     */
    apr_pool_t* getp()const{ return m_pool; }

    /**
     * get a pointer to a pointer
     */
    apr_pool_t** getpp()const{ return &m_pool; }

    /**
     * determine if this is a child (i.e. subpool)
     * @return false if not, true otherwise
     */
    bool is_child()const{
      return (apr_pool_parent_get(m_pool) ? true : false);
    }


  private:
    pool(const pool& p){}

  private:
    mutable apr_pool_t* m_pool;
    apr_status_t m_stat;
    std::string m_id;
    bool m_dtor;
  };


};
#endif
