/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#ifndef OAPR_SOCKSTREAMBUF_HPP
#define OAPR_SOCKSTREAMBUF_HPP
#include "oapr/oapr.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include <locale> // for std::codecvt
#include <streambuf>
#include "oapr/oapr_io_socket.hpp"
#include "oapr/oapr_thread_none.hpp"
FILE_VERSION("$Revision: 1.4 $",oapr_sockbuf_hpp);

namespace oapr
{

  template<typename charT,typename traits,
	   typename thread_pol_t=thread_pol_none,
	   std::streamsize bufszT=255>
  class basic_sockbuf : public std::basic_streambuf<charT,traits>
  {
  public:
    enum sockmodes{client=1,server=2,none=4};
    //standard typedefs
    typedef charT                     char_type;
    typedef typename traits::int_type int_type;
    typedef typename traits::pos_type pos_type;
    typedef typename traits::off_type off_type;
    typedef traits                    traits_type;
    typedef thread_pol_t              thread_pol_type;

    typedef std::basic_streambuf<char_type, traits_type> streambuf_type;
    typedef basic_sockbuf<char_type,traits_type,
			  thread_pol_type,bufszT> sockbuf_type;
    typedef basic_socket                                 sock_type;
    typedef typename traits_type::state_type             state_type;
    typedef std::codecvt<char_type, char, state_type>    codecvt_type;
    typedef typename codecvt_type::result                res_type;
    typedef std::ctype<char_type>                        ctype_type;

    typedef pool_mgr::pool_t pool_t;
    typedef pool_mgr::pool_ptr pool_ptr;

    friend class ios_base; // For sync_with_stdio.

  private:
    static const std::streamsize m_bufsz=bufszT;
    static const std::streamsize m_pbsize;//=bufszT/4;
    sock_type  m_sock; /**< the actual socket that we use*/
    char_type m_inbuf[bufszT];/**< buffer used during input operations*/
    char_type m_outbuf[bufszT];/**< buffer used during output operations*/
  protected:
    basic_sockbuf(sock_type &s)
      : m_sock(s){
      init_buffer();
    }

  public:
    //ctor/dtors
    basic_sockbuf(const pool_ptr&p)
      :m_sock(p)
    {
      init_buffer();
    }
    virtual ~basic_sockbuf()
    {
      /**\todo make sync on dtor a policy*/
      if(m_sock.is_open() && m_sock.is_created()){
	sync();
      }
    }
    pool_ptr gpool()const{return m_sock.gpool();}
    //extra funcs specific to socket
    /**\todo determine if apr types should be used here*/
    int create(int family, int type){return m_sock.create(family,type);}
    int connect(const sockinfo& si){return m_sock.connect(si);}
    int bind(const sockinfo& si){return m_sock.bind(si);}
    int listen(int backlog){return m_sock.listen(backlog);}
    int option(int opt, int on){return m_sock.option(opt,on);}
    sockbuf_type* accept(){
      sock_type rv = m_sock.accept();
      return new sockbuf_type(rv);
    }
    int close(){return m_sock.close();}
    int shutdown(apr_shutdown_how_e how){return m_sock.shutdown(how);}

    //status info
    bool is_created()const{return m_sock.is_created();}
    bool is_open()const{return m_sock.is_open();}
    int eof(){return m_sock.eof();}

    /**\todo impl the following funcs for the buffer, 
     * at least this is what basic_filebuf uses
     * - showmanyc
     * - underflow
     * - uflow
     * - pbackfail
     * - overflow
     * - setbuf
     * - seekoff
     * - seekpos
     * - sync
     * - imbue
     * - xsgetn
     * - xsputn
     */

  protected:
    /**\note not implementing showmanyc because apr does not
     * provide information on pending info
     */
    virtual std::streamsize showmanyc(){ return 1; }

    virtual int_type underflow()
    {
      if(gptr()<egptr()){
	return traits_type::to_int_type(*gptr());
      }
      if(fill_buffer()<=0){
	/**\todo error occurred during read, do some decent conditions*/
	return traits_type::eof();
      }else{
	return traits_type::to_int_type(*gptr());
      }
    }
    /**\note note implementing uflow until I figure it out*/
    virtual int_type pbackfail(int_type c)
    {
      if(gptr()!=eback()){
	gbump(-1);//bump back
	if(!traits_type::eq_int_type(c,traits_type::eof())){
	  *(gptr()) = traits_type::to_char_type(c);
	}
	return traits_type::not_eof(c);
      }else{
	return traits_type::eof();
      }
    }

    //following are output
    virtual std::streamsize xsputn(const char_type *s, std::streamsize n)
    {
      if(n<epptr()-pptr()){
	std::memcpy(pptr(),s,n*sizeof(char_type));
	pbump(n);
	return n;
      }else{
	for(std::streamsize i=0;i<n;++i){
	  if(traits_type::eq_int_type(sputc(s[i]),traits_type::eof())){
	    return i;
	  }
	}
	return n;
      }
    }
    virtual int_type overflow(int_type c=traits_type::eof())
    {
      if(dump_buffer()<0){
	return traits_type::eof();
      }else{
	if(!traits_type::eq_int_type(c,traits_type::eof())){
	  return sputc(c);
	}else{
	  return traits_type::not_eof(c);
	}
      }
    }
    virtual int sync(){
      return dump_buffer();
    }
  private:
    void init_buffer(){
      static char_type* initpos=m_inbuf+sockbuf_type::m_pbsize;
      setg(initpos,initpos,initpos);
      setp(m_outbuf,m_outbuf+bufszT);
    }
    int fill_buffer(){
      /**\bug need to handle exceptions here */
      //get the number to to putback
      std::streamsize npb = std::min(gptr()-eback(),sockbuf_type::m_pbsize);

      //move the bytes around
      std::memmove(m_inbuf+((sockbuf_type::m_pbsize-npb)*sizeof(char_type)),
		   gptr()-npb*sizeof(char_type),
		   npb*sizeof(char_type));

      int rrv=m_sock.read(m_inbuf+(sockbuf_type::m_pbsize*sizeof(char_type)),
			  sockbuf_type::m_bufsz-sockbuf_type::m_pbsize);
      //resetting the buffer to null if failed
      if(rrv<=0){
	/**\todo determine the validity of nulling out the buffer*/
	setg(0,0,0);
	return rrv;
      }else{
	setg(m_inbuf+sockbuf_type::m_pbsize-npb,
	     m_inbuf+sockbuf_type::m_pbsize,
	     m_inbuf+sockbuf_type::m_pbsize+rrv);
	return rrv;
      }
    }
    int dump_buffer()
    {
      /**\todo the rv is actually bytes written not status*/
      int bs=pptr()-pbase();
      int rv=m_sock.write(m_outbuf,bs);
      pbump(-bs);
      return rv;
    }
  };


  template<typename charT,typename traits,typename thread_pol_t,int bufszT>
  const std::streamsize basic_sockbuf<charT,traits,thread_pol_t,bufszT>::m_pbsize=bufszT/4;
};
#endif
